import textwrap
import yaml

from ..exception import KookiException


class MissingJar(KookiException):

    message = textwrap.dedent('''\
        Cannot find jar '{}'.
        Is it installed ? Try to run kooki bake with --update.
        Please check the jar of your config file.''')

    def __init__(self, jar):
        message = self.message.format(jar)
        super().__init__(message)


class MissingResource(KookiException):

    message = textwrap.dedent('''\
        Cannot find {0} '{1}' in the following directory:
        {2}
        Please check the {0} of your config file.''')

    def __init__(self, jars, resource_name, resource):
        from .jars import get_search_paths
        directories = get_search_paths(jars)
        directories_str = '\n'.join(['- {}'.format(directory) for directory in directories])
        message = self.message.format(resource_name, resource, directories_str)
        super().__init__(message)


class MissingFile(MissingResource):

    def __init__(self, jars, file_path):
        super().__init__(jars, 'file', file_path)


class MissingTemplate(MissingResource):

    def __init__(self, jars, template):
        super().__init__(jars, 'template', template)


class MissingMetadata(MissingResource):

    def __init__(self, jars, metadata):
        super().__init__(jars, 'metadata', metadata)


class MissingContent(MissingResource):

    def __init__(self, jars, content):
        super().__init__(jars, 'content', content)


class TooMuchResult(KookiException):

    message = textwrap.dedent('''\
        Too much files match the name provided ({})
        {}
        You should provide the full name with the file extension.''')

    def __init__(self, name, result):
        message = self.message.format(name, yaml.dump(result, default_flow_style=False)[:-1])
        super().__init__(message)


class KookiOutdated(KookiException):

    message = textwrap.dedent('''\
        Your kooki version is outdated.
        The document need \'kooki {}\', you currently have \'kooki {}\' installed.
        You should update kooki to the newer version with the following command.
        pip3 install kooki --upgrade''')

    def __init__(self, current_version, version_needed):
        message = self.message.format(version_needed, current_version)
        super().__init__(message)


class KookiJarOutdated(KookiException):

    message = textwrap.dedent('''\
        Your kooki version is outdated.
        The jar \'{}\' need \'kooki {}\', you currently have \'kooki {}\' installed.
        You should update kooki to the newer version with the following command.
        pip3 install kooki --upgrade''')

    def __init__(self, jar_name, current_version, version_needed):
        message = self.message.format(jar_name, version_needed, current_version)
        super().__init__(message)


class ErrorEvaluatingExpression(KookiException):

    def __init__(self, expression, message):
        super().__init__('')
        self.expression = []
        self.path = []
        self.message = message
        self.expression.append(expression)

    def add_expression(self, expression):
        self.expression.append(expression)

    def add_path(self, path):
        self.path.append(path)

    def __str__(self):
        message = ''
        for expression, path in zip(self.expression, self.path):
            message += expression
            message += ' in ' + path
            message += '\n'
        message += self.message
        return message



import os, yaml, toml, json, munch


class Metadata(munch.Munch):

    def update(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, dict):
                super().update(arg)
            else:
                raise Exception('arg provided is not a dict')
        super().update(**kwargs)

    def copy(self):
        copy = Metadata()
        for key, value in self.items():
            if isinstance(value, Metadata):
                copy[key] = self._copy(value)
            elif isinstance(value, list):
                new_value = []
                for i in value:
                    new_value.append(self._copy(i))
                copy[key]
            else:
                copy[key] = value
        return copy

    def _copy(self, data):
        copy = Metadata()
        for key, value in data.items():
            if isinstance(value, Metadata):
                copy[key] = self._copy(value)
            else:
                copy[key] = value
        return copy


def parse_metadata(document_metadata):
    metadata = Metadata()
    for file_path, file_content in document_metadata.items():
        new_metadata = load(file_path, file_content)
        metadata = data_merge(new_metadata, metadata)
    return Metadata.fromDict(metadata)


def load(file_path, file_content):

    root, ext = os.path.splitext(file_path)

    if ext == '.yaml' or ext == '.yml':
        metadata = yaml.safe_load(file_content)
    elif ext == '.toml':
        metadata = toml.loads(file_content)
    elif ext == '.json':
        metadata = json.loads(file_content)
    else:
        raise Exception('No loader for this type of file \'{}\''.format(ext))

    return metadata


def data_merge(a, b):

    class MergeError(Exception):
        pass

    key = None

    try:
        if a is None or isinstance(a, str) or isinstance(a, bytes) \
           or isinstance(a, int) or isinstance(a, float):
            a = b
        elif isinstance(a, list):
            if isinstance(b, list):
                a.extend(b)
            else:
                a.append(b)
        elif isinstance(a, dict):
            if isinstance(b, dict):
                for key in b:
                    if key in a:
                        a[key] = data_merge(a[key], b[key])
                    else:
                        a[key] = b[key]
            else:
                raise MergeError('Cannot merge non-dict "%s" into dict "%s"' % (b, a))
        else:
            raise MergeError('NOT IMPLEMENTED "%s" into "%s"' % (b, a))
    except TypeError as e:
        raise MergeError('TypeError "%s" in key "%s" when merging "%s" into "%s"' % (e, key, b, a))
    return a



import pretty_output, os, yaml

from collections import OrderedDict
from packaging import version

from kooki.tools import read_file

from .jars import search_file, search_jar, jar_dependencies
from .exception import MissingTemplate, MissingMetadata, MissingJar
from .exception import MissingContent, KookiException, KookiOutdated
from .metadata import Metadata, parse_metadata
from .common import apply_template, export_to
from .extension import Extension, Caller
from ..version import __version__

output = []


def output_search_start():
    global output
    output = []


def output_search(path, fullpath):
    if fullpath:
        output.append({'name': path, 'status': '[found]', 'path': fullpath})
    else:
        output.append({'name': path, 'status': ('[missing]', 'red'), 'path': ''})


def output_search_finish():
    pretty_output.infos(output, [('name', 'blue'), ('status', 'green'), ('path', 'cyan')])


def process_document(document, step=lambda *args, **kwargs: None):
    # jars
    step('jars')
    load_jars(document)

    # version
    check_kooki_version(document.version)

    # template
    step('template')
    template_path, template_extension = load_template(document)
    document.extension = template_extension
    template = Extension(document, template_path)

    # metadata
    step('metadata')
    load_metadata(document)

    # process
    step('processor')
    document_content = template()
    output = apply_template(document.output, document.metadata)
    output_path = os.path.join(document.context, output.replace('\n', ''))
    Caller.clean()

    # output
    step('output')
    absolute_file_path = export_to(output_path, template_extension, document_content)
    pretty_output._print_colored('export to ', 'blue', '')
    pretty_output._print_colored(absolute_file_path, 'cyan')

    # toppings
    step('toppings')
    import subprocess
    os.environ['KOOKI_CONTEXT'] = document.context
    os.environ['KOOKI_EXPORT_FILE'] = absolute_file_path
    for topping in document.toppings:
        pretty_output._print_colored('execute ', 'blue', '')
        pretty_output._print_colored(topping, 'cyan')
        subprocess.call(topping, shell=True)
        pretty_output._print_colored('finish ', 'blue', '')
        pretty_output._print_colored(topping, 'cyan')


def check_kooki_version(kooki_version_needed):
    if version.parse(__version__) < version.parse(kooki_version_needed):
        raise KookiOutdated(__version__, kooki_version_needed)


def load_template(document):
    output_search_start()
    file_full_path = search_file(document, document.template)
    output_search(document.template, file_full_path)
    if file_full_path:
        file_read = read_file(file_full_path)
        document.template = file_read
        output_search_finish()
    else:
        output_search_finish()
        raise MissingTemplate(document.jars, document.template)

    _, template_extension = os.path.splitext(file_full_path)
    return file_full_path, template_extension


def print_extension(extensions, prefix=''):
    extensions_name = list(extensions.keys())
    extensions_name.sort()
    if prefix != '':
        prefix += '.'
    for extension_name in extensions_name:
        extension = extensions[extension_name]
        if isinstance(extension, dict):
            print_extension(extension, extension_name)
        else:
            output_search('{}{}'.format(prefix, extension_name), extension.path)


def load_jars(document):
    full_path_jars = []
    output_search_start()
    for jar in document.jars:
        load_jar_dependencie(full_path_jars, jar)
        jar_full_path = search_jar(jar)
        if jar_full_path:
            output_search(jar, jar_full_path)
            full_path_jars.append(jar_full_path)
        else:
            output_search_finish()
            raise MissingJar(jar)
    document.jars = full_path_jars
    output_search_finish()


def load_jar_dependencie(full_path_jars, jar):
    dependencies = jar_dependencies(jar)
    for dependencie in dependencies:
        load_jar_dependencie(full_path_jars, dependencie)
        jar_full_path = search_jar(dependencie)
        if jar_full_path:
            output_search(dependencie, jar_full_path)
            full_path_jars.append(jar_full_path)
        else:
            output_search_finish()
            raise MissingJar(jar)


def load_metadata(document):
    metadata_full_path = {}
    output_search_start()
    for metadata in document.metadata:
        file_full_path = search_file(document, metadata)
        output_search(metadata, file_full_path)
        if file_full_path:
            file_read = read_file(file_full_path)
            metadata_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingMetadata(document.jars, metadata)
    document.metadata = metadata_full_path
    output_search_finish()
    document.metadata = parse_metadata(document.metadata)

    pretty_output.debug_on()
    pretty_output.colored(yaml.dump(document.metadata, default_flow_style=False), 'white', 'on_yellow')
    pretty_output.debug_off()


def print_metadata(metadata, prefix=''):
    matadata_keys = list(metadata.keys())
    matadata_keys.sort()
    if prefix != '':
        prefix += '.'
    for matadata_key in matadata_keys:
        element = metadata[matadata_key]
        if isinstance(element, dict):
            print_metadata(element, matadata_key)
        else:
            output_search('{}{}'.format(prefix, matadata_key), '')



import os
import shlex
import struct
import logging
import platform
import subprocess
import termcolor



logger = logging.getLogger()
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
formatter = logging.Formatter('%(message)s')
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.INFO)
logger.addHandler(stream_handler)

_color_policy = True
_output_policy = True
_debug_policy = False
_debug = False
_message = ''


def _get_width():
    result = get_terminal_size()
    return result[0]


def debug_on():
    global _debug
    _debug = True


def debug_off():
    global _debug
    _debug = False


def output_control(func):
    def wrapper(*args, **kwargs):
        global _output_policy
        if _output_policy:
            func(*args, **kwargs)
    return wrapper


def debug_control(func):
    def wrapper(*args, **kwargs):
        global _debug_policy, _debug
        if not _debug or (_debug and _debug_policy):
            func(*args, **kwargs)
    return wrapper


def set_color_policy(color_policy):
    global _color_policy
    _color_policy = color_policy


def set_output_policy(output_policy):
    global _output_policy
    _output_policy = output_policy


def set_debug_policy(debug):
    global _debug_policy
    _debug_policy = debug


@output_control
def color(full_text):
    results = full_text.split('[')
    for result in results:
        if result and result != '':
            second_results = result.split(']')
            color, text = second_results[0], second_results[1]
            _print_colored(text, color, end='')


@output_control
def title_1(name=None):
    if name:
        _step(name, 'white', '=', '[', ']')
    else:
        _step('', 'white', '=')


@output_control
def title_2(name=None):
    if name:
        _step(name, 'white', '-', '[', ']')
    else:
        _step('', 'white', '-')


@output_control
def title_3(name=None):
    if name:
        _step(name, 'yellow', '-')
    else:
        _step('', 'yellow', '-')


@output_control
def title_4(name=None):
    if name:
        _step(name, 'magenta', '-')
    else:
        _step('', 'green', '-')


@output_control
def start_step(name):
    _step(name, 'yellow', '-')


@output_control
def error_step(name):
    _step(name, 'red', '-')


@output_control
def category(message):
    _print_colored('[' + message + ']', 'magenta')


def error(message):
    _print_colored(message, 'red')


@output_control
def warning(message):
    _print_colored(message, 'orange')


@output_control
def info(message):
    _print_colored(message, 'cyan')


def _step(name, text_color, character, before='', after=''):
    width = _get_width()
    if name != '':
        name = ' {} '.format(name)
    text = '{}{}{}'.format(before, name, after)
    _print_colored(text.center(width, character), text_color)


@output_control
def push(message):
    pass


@output_control
def infos(infos, attrs):
    text = ''
    spaces = {}
    shift = 2

    for info in infos:
        for attr in attrs:
            name = attr[0]
            info_name_len = len(str(info[name]))
            if name in spaces:
                if info_name_len + shift > spaces[name]:
                    spaces[name] = info_name_len + shift
            else:
                spaces[name] = info_name_len + shift

    for info in infos:
        for attr in attrs:
            name = attr[0]
            color = attr[1]
            text = ''
            if len(attr) > 2:
                text = attr[2]

            if isinstance(info[name], tuple):
                info_name = str(info[name][0])
                info_name_len = len(info_name)
                full_text = text + info_name
                color = info[name][1]
            else:
                info_name = str(info[name])
                info_name_len = len(info_name)
                full_text = text + info_name

            _print_colored(full_text, color, end='')
            _print_colored(' ' * (spaces[name] - info_name_len), end='')
        _print_colored()


@output_control
def clear():
    pass


def colored(*args, **kwargs):
    _print_colored(*args, **kwargs)

@output_control
@debug_control
def _print_colored(text='', color=None, highlight=None, end='\n'):
    global _message
    global _color_policy

    if _color_policy:

        if highlight and color:
            _message += termcolor.colored(text, color, highlight)
        elif color:
            _message += termcolor.colored(text, color)
        else:
            _message += termcolor.colored(text)
    else:
        _message += text

    if end == '\n':
        logger.info(_message)
        _message = ''


def get_terminal_size():
    """ getTerminalSize()
     - get width and height of console
     - works on linux,os x,windows,cygwin(windows)
     originally retrieved from:
     http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    """
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        print('default')
        tuple_xy = (80, 25)      # default value
    return tuple_xy


def _get_terminal_size_windows():
    try:
        from ctypes import windll, create_string_buffer
        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            (bufx, bufy, curx, cury, wattr,
             left, top, right, bottom,
             maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            sizex = right - left + 1
            sizey = bottom - top + 1
            return sizex, sizey
    except:
        pass


def _get_terminal_size_tput():
    try:
        cols = int(subprocess.check_call(shlex.split('tput cols')))
        rows = int(subprocess.check_call(shlex.split('tput lines')))
        return (cols, rows)
    except:
        pass


def _get_terminal_size_linux():
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            cr = struct.unpack('hh',
                               fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            return cr
        except:
            pass
    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (os.environ['LINES'], os.environ['COLUMNS'])
        except:
            return None
    return int(cr[1]), int(cr[0])


import empy, os
import pretty_output

from collections import OrderedDict

from kooki.tools import get_front_matter, write_file

from .metadata import Metadata
from .exception import ErrorEvaluatingExpression


def export_to(name, extension, content):
    file_name = '{0}{1}'.format(name, extension)
    write_file(file_name, content)
    absolute_path = os.path.join(os.getcwd(), file_name)
    return absolute_path


def apply_template(data, metadata, prefix='@'):
    result = ''
    front_matter, content = get_front_matter(data)
    unique_metadata = get_metadata(front_matter, metadata)
    result = apply_interpreter(content, unique_metadata, prefix)
    return result


def get_metadata(front_matter, metadata):
    metadata_copy = Metadata()
    metadata_copy.update(front_matter)
    metadata_copy.update(metadata)
    return metadata_copy


def apply_interpreter(content, metadata, prefix):
    interpreter = empy.Interpreter()
    interpreter.setPrefix(prefix)
    result = interpreter.expand(content, metadata)
    return result


from cerberus import Validator
from .document import Document

document_format = Document.format()
document_fields = Document.fields()


document_version_1 = {
    **document_format,
    'default': {
        'required': False,
        'type': 'dict',
        'excludes': document_fields,
        'dependencies': 'kooki',
        'schema': document_format
    },
    'kooki': {
        'required': False,
        'type': 'dict',
        'excludes': document_fields,
        'keyschema': {'type': 'string', 'regex': '[A-Za-z0-9_\-]+'},
    }
}


def check_config_format(config):
    v = Validator(document_version_1)
    config_valid = v.validate(config, document_version_1)
    if not config_valid:
        raise Exception(v.errors)



import re
from mistune import Markdown, Renderer, InlineGrammar, InlineLexer
from munch import Munch

def render(content, extensions):
    renderer = KookiRenderer(extensions)
    markdown = KookiMarkdown(renderer=renderer)
    return markdown(content)


class KookiInlineGrammar(InlineGrammar):

    criticmarkup_deletion = re.compile(r'^{--(?=\S)([\s\S]*?\S)--}')
    criticmarkup_substitution = re.compile(r'^{~~(?=\S)([\s\S]*?\S)~>(?=\S)([\s\S]*?\S)~~}')
    criticmarkup_comment = re.compile(r'^{>>(?=\S)([\s\S]*?\S)<<}')
    criticmarkup_highlight = re.compile(r'^{==(?=\S)([\s\S]*?\S)==}{>>(?=\S)([\s\S]*?\S)<<}')
    criticmarkup_addition = re.compile(r'^{\+\+(?=\S)([\s\S]*?\S)\+\+}')
    text = re.compile(r'^[\s\S]+?(?=[\\<!\[_*`~{}]|https?://| {2,}\n|$)')


class KookiInlineLexer(InlineLexer):

    default_rules = [
        'escape', 'inline_html', 'autolink', 'url',
        'footnote', 'link', 'reflink', 'nolink',
        'double_emphasis', 'emphasis', 'code',
        'criticmarkup_addition', 'criticmarkup_deletion',
        'criticmarkup_substitution', 'criticmarkup_comment',
        'criticmarkup_highlight', 'linebreak',
        'strikethrough', 'text',
    ]

    def __init__(self, renderer, **kwargs):
        rules = KookiInlineGrammar()
        super(KookiInlineLexer, self).__init__(renderer, rules, **kwargs)

    def output_criticmarkup_addition(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_addition(text)

    def output_criticmarkup_deletion(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_deletion(text)

    def output_criticmarkup_substitution(self, m):
        text = m.group(1)
        substitution = m.group(2)
        return self.renderer.criticmarkup_substitution(text, substitution)

    def output_criticmarkup_comment(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_comment(text)

    def output_criticmarkup_highlight(self, m):
        text = m.group(1)
        comment = m.group(2)
        return self.renderer.criticmarkup_highlight(text, comment)


class KookiMarkdown(Markdown):

    def __init__(self, renderer, **kwargs):
        if 'inline' not in kwargs:
            kwargs['inline'] = KookiInlineLexer(renderer)
        super(KookiMarkdown, self).__init__(renderer, **kwargs)


def catch_errors(func):
    def callback(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return ''
    return callback


class KookiRenderer(Renderer):
    
    data = Munch()
    
    @classmethod
    def clean(cls):
        cls.data = Munch()

    def __init__(self, extensions, **kwargs):
        self.extensions = extensions
        super(KookiRenderer, self).__init__(**kwargs)

    def block_quote(self, text):
        return self.extensions.__block_quote__(renderer=KookiRenderer.data, text=text)

    def block_html(self, html):
        return self.extensions.__block_html__(renderer=KookiRenderer.data, html=html)

    def inline_html(self, html):
        return self.extensions.__inline_html__(renderer=KookiRenderer.data, html=html)

    def header(self, text, level, raw):
        return self.extensions.__header__(renderer=KookiRenderer.data, text=text, level=level, raw=raw)

    def hrule(self):
        return self.extensions.__hrule__(renderer=KookiRenderer.data)

    def list(self, body, ordered):
        return self.extensions.__list__(renderer=KookiRenderer.data, body=body, ordered=ordered)

    def list_item(self, text):
        return self.extensions.__list_item__(renderer=KookiRenderer.data, text=text)

    def paragraph(self, text):
        return '{}\n\n'.format(self.extensions.__paragraph__(renderer=KookiRenderer.data, text=text))

    def table(self, header, body):
        return self.extensions.__table__(renderer=KookiRenderer.data, header=header, body=body)

    def table_row(self, content):
        return self.extensions.__table_row__(renderer=KookiRenderer.data, content=content, placeholder=False)

    def table_cell(self, content, **flags):
        return self.extensions.__table_cell__(renderer=KookiRenderer.data, content=content, placeholder=False, **flags)

    def link(self, link, title, content):
        return self.extensions.__link__(renderer=KookiRenderer.data, link=link, title=title, content=content)

    def autolink(self, link, is_email=False):
        return self.extensions.__autolink__(renderer=KookiRenderer.data, link=link, is_email=is_email)

    def block_code(self, code, language):
        return self.extensions.__block_code__(renderer=KookiRenderer.data, code=code, language=language)

    def codespan(self, text):
        return self.extensions.__codespan__(renderer=KookiRenderer.data, text=text)

    def double_emphasis(self, text):
        return self.extensions.__double_emphasis__(renderer=KookiRenderer.data, text=text)

    def emphasis(self, text):
        return self.extensions.__emphasis__(renderer=KookiRenderer.data, text=text)

    def image(self, src, title, alt_text):
        return self.extensions.__image__(renderer=KookiRenderer.data, src=src, title=title, alt_text=alt_text)

    def strikethrough(self, text):
        return self.extensions.__strikethrough__(renderer=KookiRenderer.data, text=text)

    def text(self, text):
        return self.extensions.__text__(renderer=KookiRenderer.data, text=text)

    def linebreak(self):
        return self.extensions.__linebreak__(renderer=KookiRenderer.data)

    def newline(self):
        return self.extensions.__newline__(renderer=KookiRenderer.data)

    def footnote_ref(self, key, index):
        return self.extensions.__footnote_ref__(renderer=KookiRenderer.data, key=key, index=index)

    def footnote_item(self, key, text):
        return self.extensions.__footnote_item__(renderer=KookiRenderer.data, key=key, text=text)

    def footnotes(self, text):
        return self.extensions.__footnotes__(renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_addition(self, text):
        return self.extensions.__criticmarkup_addition__(renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_deletion(self, text):
        return self.extensions.__criticmarkup_deletion__(renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_substitution(self, text, substitution):
        return self.extensions.__criticmarkup_substitution__(renderer=KookiRenderer.data, text=text, substitution=substitution)

    @catch_errors
    def criticmarkup_comment(self, text):
        return self.extensions.__criticmarkup_comment__(renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_highlight(self, text, comment):
        return self.extensions.__criticmarkup_highlight__(renderer=KookiRenderer.data, text=text, comment=comment)



from .check import check_config_format
from .document import Document


def parse_document_rules(config):
    check_config_format(config)
    if 'kooki' in config:
        return parse_multi_documents(config)
    else:
        return parse_mono_document(config)


def parse_mono_document(config):
    document = Document(config)
    return document


def parse_multi_documents(config):
    documents = {}
    for document_name, document_config in config['kooki'].items():

        if 'default' in config:
            if document_config:
                if 'output' not in config['default'] and 'output' not in document_config:
                    document_config['output'] = document_name
                if 'context' not in config['default'] and 'context' not in document_config:
                    document_config['context'] = document_name
            else:
                document_config = {}
                document_config['output'] = document_name
                document_config['context'] = document_name
            document = Document(config['default'], document_config)

        else:
            if document_config:
                if 'output' not in document_config:
                    document_config['output'] = document_name
                if 'context' not in document_config:
                    document_config['context'] = document_name
            else:
                document_config = {}
                document_config['output'] = document_name
                document_config['context'] = document_name
            document = Document(document_config)

        documents[document_name] = document
    return documents



import os, empy, tempfile, textwrap
import pretty_output
import random, string

from kooki.tools import get_front_matter, read_file

from .jars import search_file, search_file_in_context
from .common import apply_template
from .metadata import Metadata
from .renderer import render
from .exception import ErrorEvaluatingExpression, MissingFile

PREFIX = '@'
MARKDOWN_EXT = '.md'


class Extension():

    level = 0

    def __init__(self, document, extension_path):
        file_content = read_file(extension_path)
        _, self.file_extension = os.path.splitext(extension_path)
        self.extension_path = extension_path
        self.document = document
        self.front_matter, self.content = get_front_matter(file_content)

    def __call__(self, *args, **kwargs):

        try:
            metadata = self.parse_metadata(*args, **kwargs)
            content = self.content

            self.debug_tree_start()

            caller = Caller(self.document, self.extension_path, metadata=metadata, file_extension=self.file_extension)
            content = self.apply_interpreter(content, caller, prefix=PREFIX)

            if self.file_extension == MARKDOWN_EXT:
                content = render(content, caller)

            for placeholder_key, placeholder_value in Caller.placeholders.items():
                content = content.replace(placeholder_key, placeholder_value)

            self.debug_tree_stop()

            return content

        except ErrorEvaluatingExpression as e:
            e.add_path(self.extension_path)
            raise e

    def apply_interpreter(self, content, metadata, prefix):
        interpreter = empy.Interpreter()
        interpreter.setPrefix(prefix)
        result = interpreter.expand(content, metadata)
        return result

    def parse_metadata(self, *args, **kwargs):
        metadata = Metadata()
        metadata.update(self.front_matter)
        metadata.update(self.document.metadata)
        new_args = {}
        for index, arg in enumerate(args):
            new_args['arg{}'.format(index)] = arg
        metadata.update(**new_args)
        metadata.update(**kwargs)
        return metadata

    def debug_tree_start(self):
        pretty_output.debug_on()
        if Extension.level == 0:
            pretty_output.colored(self.extension_path, 'white', 'on_yellow')
        else:
            message = ''
            for i in range(0, self.level-1):
                message += '│   '
            message += '├'
            message += '─' * 2
            message += ' '
            pretty_output.colored(message, 'white', 'on_yellow', end='')
            pretty_output.colored(self.extension_path, 'white', 'on_yellow')
        Extension.level += 1
        char = '└'
        pretty_output.debug_off()

    def debug_tree_stop(self):
        Extension.level -= 1


def is_builtin_class_instance(obj):
    return obj.__class__.__module__ == '__builtin__'


class Caller():

    placeholders = {}
    tempary_extensions = []

    @classmethod
    def clean(cls):
        for tempary_extension in cls.tempary_extensions:
            os.unlink(tempary_extension.name)

    def __init__(self, document, extension_path, metadata, file_extension, call_path=''):
        self.document = document
        self.extension_path = extension_path
        self.call_path = call_path
        self.metadata = metadata
        import pretty_output
        self.file_extension = file_extension
        self.metadata.update({
            'find': self.find,
            'load': self.load,
            'get': self.get})
        self.data = None

    def __getattr__(self, name):
        return self.get_item(name)

    def __getitem__(self, key):
        return self.get_item(key)

    def __setitem__(self, key, value):
        self.metadata[key] = value

    def get_item(self, key):

        if key == 'locals':
            def call_locals():
                return locals()
            return call_locals

        if self.call_path != '':
            new_call_path = '{}/{}'.format(self.call_path, key)
        else:
            new_call_path = key

        self.metadata.update(__builtins__)
        self.metadata.update(locals()['self'].metadata)
        value = self.metadata.get(key)

        if value != None:
            if is_builtin_class_instance(value):
                return type('', (type(value), Caller), {})(self.document, self.extension_path, self.metadata, file_extension=self.file_extension, call_path=new_call_path)
            else:
                return value
        else:
            return Caller(self.document, self.extension_path, self.metadata, file_extension=self.file_extension, call_path=new_call_path)

    def __repr__(self):
        return ''

    def __call__(self, *args, placeholder=True, **kwargs):

        path = search_file_in_context(self.document, self.extension_path, self.call_path)
        if not path:
            path = search_file(self.document, self.call_path)

        if not path:
            raise Exception('cannot find {}'.format(self.call_path))

        if os.path.isdir(path):
            full_path_md = os.path.join(path, '__kooki__.md')
            full_path_spec = os.path.join(path, '__kooki__{}'.format(self.file_extension))
            if os.path.isfile(full_path_md):
                full_path = full_path_md
            elif os.path.isfile(full_path_spec):
                full_path = full_path_spec
            else:
                raise Exception(textwrap.dedent('''\
                    You try to call this extension: {}
                    Cannot find one of those files in the directory.
                    - {}
                    - {}
                    '''.format(path, full_path_md, full_path_spec)))
        else:
            full_path = path

        extension = Extension(self.document, full_path)
        result = extension(*args, **kwargs)
        if placeholder:
            name = self.get_var_name()
            Caller.placeholders[name] = result
            return '{}'.format(name)
        else:
            return result

    def find(self, file_path):
        file_full_path = search_file(self.document, file_path)
        if file_full_path:
            return file_full_path
        else:
            raise MissingFile(self.document.jars, file_path)

    def load(self, content, template_extension):
        directory_path = os.path.dirname(self.extension_path)
        temp_extension = tempfile.NamedTemporaryFile(suffix=template_extension, dir=directory_path, delete=False)
        temp_extension.write(bytes(content, 'utf-8'))
        temp_extension.close()
        Caller.tempary_extensions.append(temp_extension)
        return Extension(self.document, temp_extension.name)

    def get(self, key, default=''):
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def get_var_name(self):
        name = ''.join(random.choice(string.ascii_uppercase) for _ in range(25))
        while name in Caller.placeholders:
            name = ''.join(random.choice(string.ascii_uppercase) for _ in range(25))
        return name

